/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "config.h"
#include "qop.h"
#include "jniint.h"
#include "log.h"
#include "dynamic_loading.h"
#include "exceptions.h"
#include "interp.h"
#include "sig.h"
#include "ClazzFile.h"
#include "nativeglue.h"
#include "compat.h"
#include "primitive-class.h"
#include "util.h"

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <assert.h>

#include "op_stack.h"

#define MYLOG "NativeGlue"

/*
 * Convert (any/class/Name, method) to Java_any_class_Name_method.
 * The returned char * is allocated off the heap using strdup and must
 * be free'ed.
 */
static char*
create_short_native_name(char *classname, MethodStruct *method)
{
  char name_buf[1024];
  char *bar_classname = strdup(classname);
  int retval;
  char *method_name = method->name;

  slashes_to_underscores(bar_classname);

  retval = snprintf(name_buf, sizeof(name_buf),
		    "Java_%s_%s_stub",
		    bar_classname, method_name);
  free(bar_classname);

  if (-1 == retval)
    { /* XXX Output was truncated.  What should we do? */
      fprintf(stderr, "Error: Class method %s.%s truncated.\n",
	      classname, method_name);
    }

  return strdup(name_buf);
}


/*
 * Convert (any/class/Name, method) to Java_any_class_Name_method__sig.
 * The returned char * is allocated off the heap using strdup and must
 * be free'ed.
 */
static char*
create_long_native_name(JNIEnv *env, char *classname,
			MethodStruct *method)
{
  char name_buf[1024];
  char *bar_classname = strdup(classname);
  int retval;
  char *method_name = method->name;
  char *hacked_sig = SIG_formatToNativeName(env, method->sig);

  slashes_to_underscores(bar_classname);

  retval = snprintf(name_buf, sizeof(name_buf),
		    "Java_%s_%s__%s_stub",
		    bar_classname, method_name, hacked_sig);
  free(bar_classname);
  free(hacked_sig);

  if (-1 == retval)
    { /* XXX Output was truncated.  What should we do? */
      fprintf(stderr, "Error: Class method %s.%s truncated.\n",
	      classname, method_name);
    }

  return strdup(name_buf);
}

static void
resolve_native_func(JNIEnv *env, MethodStruct *method)
{
  char *class_name = getClassName(env, method->clazz);
  char *native_func_name;
  
  native_func_name = create_short_native_name(class_name, method);
  JAVARLOG1(MYLOG, 1, "   short native stub name is %s()\n", native_func_name);
  
  method->native_func = (NativeFunc)DLL_findFunction(native_func_name);
  
  if (method->native_func == NULL)
    {
      /* now we try the long name. */
      free(native_func_name);
      native_func_name = create_long_native_name(env, class_name, method);

      method->native_func = (NativeFunc)DLL_findFunction(native_func_name);
    }

  if (method->native_func == NULL) /* what do we do here? */
    {
      char *msg;
      
      msg = (char*)malloc(sizeof(char) 
			  * (strlen("Unresolved Function: ")
			     + strlen(native_func_name) + 1));
      assert(NULL != msg);
      sprintf(msg, "Unresolved Function: %s", native_func_name);
      
      (*env)->FatalError(env, msg);
    }
  
  free(native_func_name);
}

static jvalue
do_native_method_call_with_args(JNIEnv *env,
				MethodStruct *method,
				jvalue *args)
{
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  GC_obj *head_obj;
  StackFrame *new_frame = push_frame(henv->_java_info,
				     0);

  new_frame->flags |= FRAME_NATIVE;
  METHOD(new_frame) = method;
#ifdef DEBUG
  CLASSNAME(new_frame) = getClassName(env, method->clazz);
  METHODNAME(new_frame) = strdup (METHOD(new_frame)->name);
  assert(NULL != METHODNAME(new_frame));
#endif

  /* get the function pointer for this native method, if we don't already
     know it. */
  if (!method->native_func)
    resolve_native_func(env, method);

  OPSTACK_TOP(new_frame) = OPSTACK(new_frame)->stack_top;

  if (method->access_flags & ACC_STATIC)
    THISPTR(new_frame) = NULL;
  else
    THISPTR(new_frame) = args[1].l;

  /* save off the stopping point on thread-local deletions */
  head_obj = henv->_local_refs;

  maybe_enter_monitor_for_method(ENV(new_frame), METHOD(new_frame), THISPTR(new_frame));

  (*METHOD(new_frame)->native_func)(env, args);

#ifdef garbage_collect_working
  /* collect all the objects that were allocated during this native call */
  {
    while (henv->_local_refs
	   && henv->_local_refs != head_obj)
      {
	GC_obj *foo = henv->_local_refs;
	GC_obj *next = foo->next;

	UNQUEUE(foo, henv->_local_refs);

	GC_deleteThreadLocal((JavaVM*)henv->_vm, foo);
	henv->_local_refs = next;
      }
  }
#endif

  if (!SIG_isVoid(env, METHOD(new_frame)->sig->method.return_type))
    switch(METHOD(new_frame)->sig->method.return_type->any.tag)
      {
      case SIG_PRIM:
	THREAD_INFO(new_frame)->return_value.tag = METHOD(new_frame)->sig->method.return_type->prim.type;
	break;
      case SIG_ARRAY:
      case SIG_CLASS:
	THREAD_INFO(new_frame)->return_value.tag = SIG_JOBJECT;
	break;
      case SIG_METHOD:
	assert(0); /* this shouldn't ever happen. */
	break;
      }

  THREAD_INFO(new_frame)->return_value.value = args[0];

  maybe_push_return_value(new_frame);
  
  /* then pop the frame off */
  pop_frame(THREAD_INFO(new_frame));

  return THREAD_INFO(new_frame)->return_value.value;
}

void
do_native_method_call(JNIEnv *env, MethodStruct *method)
{
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  int num_arguments = 0;
  jvalue args[100];
  int i;
#ifdef LOG
  char *class_name = getClassName(env, method->clazz);
#endif
  StackFrame *current_frame = TOPFRAME(henv->_java_info);

  JAVARLOG2(MYLOG, 1, "Inside do_native_method_call(%s.%s)\n",
	    class_name, method->name);

  initialize_class(env, method->clazz);

  num_arguments = SIG_numParams(env, method->sig);
  
  for (i = num_arguments + 1;
       i > 1;
       i --)
    {
      op_stack_pop_value(env, OPSTACK(current_frame),
			 (method->sig->method.params[ i - 2  ]->any.tag == SIG_PRIM
			  ? method->sig->method.params[ i - 2 ]->prim.type
			  : SIG_JOBJECT),
			 &args[ i ]);
    }
  
  if (method->access_flags & ACC_STATIC)
    {
      /* fill in the class pointer */
      args[1].l = clazzfile_to_jclass(env, method->clazz);
    }
  else
    {
      /* pop the this pointer */
      op_stack_pop_value(env, OPSTACK(current_frame),
			 SIG_JOBJECT,
			 &args[ 1 ]);
    }

  do_native_method_call_with_args(env, method, args);
}

/* This is the generic method used by the two specific below. */
static jvalue
call_java_method(JNIEnv *env,
		 MethodStruct *method,
		 jobject obj, /* NULL for static calls */
		 jvalue *args)
{
  StackFrame *new_frame;
  HungryJNIEnv *henv = (HungryJNIEnv*)env;

  JAVARLOG2(MYLOG, 4, "calling %s.%s()\n",
	    method->clazz->class_name,
	    method->name);

  initialize_class(env, method->clazz);

  new_frame = create_frame_for_method(henv->_java_info, method);
  if (NULL == new_frame)
    {
      jvalue value;
      /*
       * Fill it with zero.  Assume jobject is the longest type.  I'm
       * not sure what this method should return on error, but...
       */
      value.l = 0;
      return value; /* create_frame_for_methods has already throwed */
    }

  fill_local_vars(new_frame, method, args, obj);

  OPSTACK_TOP(new_frame) = OPSTACK(new_frame)->stack_top;

  THREAD_INFO(new_frame)->return_value.value.j = 0;

  maybe_enter_monitor_for_method(env, method, obj);

  interp_loop(new_frame);

  OPSTACK(new_frame)->stack_top = OPSTACK_TOP(new_frame);

  return THREAD_INFO(new_frame)->return_value.value;
}
  
jvalue
CallJavaMethod (JNIEnv *env,
		MethodStruct *method,
		jobject obj,
		jvalue *args)
{
  japhar_obj real_obj = cast_obj((japhar_obj)obj,
				 method->clazz);
  assert(real_obj);
  if(!real_obj)
    {
      /* this should be an exception, but I don't know what... */
      jvalue foo;

      foo.l = 0;
      return foo;
    }

  if (method->access_flags & ACC_NATIVE)
    {
      jvalue new_args[100];
      int num_arguments = SIG_numParams(env, method->sig) + 1;
      int i;

      for (i = num_arguments;
	   i > 1;
	   i --)
	{
	  new_args[i] = args[i - 1];
	}

      new_args[1].l = real_obj;

      return do_native_method_call_with_args(env,
					     method, new_args);
    }
  else
    {
      return call_java_method(env, method, real_obj, args);
    }
}

jvalue
CallStaticJavaMethod (JNIEnv *env,
		      MethodStruct *method,
		      jvalue *args)
{
  if (method->access_flags & ACC_NATIVE)
    {
      jvalue new_args[100];
      int num_arguments = SIG_numParams(env, method->sig) + 1;
      int i;

      for (i = num_arguments;
	   i >=  2;
	   i --)
	{
	  new_args[i] = args[i - 1];
	}

      new_args[1].l = 0;

      return do_native_method_call_with_args(env,
					     method, new_args);
    }
  else
    {
      return call_java_method(env, method, 0, args);
    }
}

/****************************************************************************
 * These are used by libjni/src/jninativ.c to register native method
 * function pointers.
 */
void
register_native_func(JNIEnv *env, MethodStruct *method, void *fnPtr)
{
  /*
   * XXX Should probably save the old function to be able to restore
   * previous state
   */
  method->native_func = fnPtr;
}
void
unregister_native_func(JNIEnv *env, MethodStruct *method)
{
  /*
   * XXX Should probably restore to old value
   */
  method->native_func = NULL;
}

jclass
sig_to_jclass(JNIEnv *env, Signature *sig)
{
  jclass sig_cls = NULL;

  if (sig->any.tag == SIG_CLASS
      || sig->any.tag == SIG_ARRAY)
    {
      char *java_sig;
	  java_sig = SIG_formatToJavaSig(env, sig);
	  
      sig_cls = (*env)->FindClass(env, java_sig);
	  
      free(java_sig);
    }
  else /* SIG_PRIM */
    {
      switch (sig->prim.type)
	{
	case SIG_JBOOLEAN:
	  sig_cls = createFakePrimitiveClass(env, "boolean");
	  break;
	case SIG_JBYTE:
	  sig_cls = createFakePrimitiveClass(env, "byte");
	  break;
	case SIG_JCHAR:
	  sig_cls = createFakePrimitiveClass(env, "char");
	  break;
	case SIG_JSHORT:
	  sig_cls = createFakePrimitiveClass(env, "short");
	  break;
	case SIG_JINT:
	  sig_cls = createFakePrimitiveClass(env, "int");
	  break;
	case SIG_JLONG:
	  sig_cls = createFakePrimitiveClass(env, "long");
	  break;
	case SIG_JFLOAT:
	  sig_cls = createFakePrimitiveClass(env, "float");
	  break;
	case SIG_JDOUBLE:
	  sig_cls = createFakePrimitiveClass(env, "double");
	  break;
	case SIG_JVOID:
	  sig_cls = createFakePrimitiveClass(env, "void");
	  break;
	default:
	  (*env)->FatalError(env, "Illegal primitive type.");
	  break;
	}
    }

  return sig_cls;
}

/**
 * Convert from a methodID into a Java object of class
 * java.lang.reflect.Method or java.lang.reflect.Constructor
 * (whichever corresponds).  The only difference with the JDK 1.2 JNI
 * ToReflectedMethod function is that by using this method,
 * repetitive string lookups can be avoided when inside loops.
 */
jobject
to_reflected_method(JNIEnv *env, jmethodID methodID,
    jclass class_class, jclass method_class, jmethodID method_ctor,
    jfieldID clazz_field, jfieldID slot_field, jfieldID name_field,
    jfieldID returnType_field, jfieldID parameterTypes_field,
    jfieldID exceptionTypes_field)
{
  MethodStruct *method = (MethodStruct *) methodID;
  ClazzFile *clazz = method->clazz;
  jobject rmethod = (*env)->NewObject(env, method_class, method_ctor);
  int num_params = method->sig->method.num_params;
  int num_exceptions = method->num_throwable_exceptions;
  jobjectArray exception_array;
  jobjectArray param_array;
  int p;
  int slot;

  /* Determine method's slot inside class */
  for (slot = 0; slot < clazz->num_methods; ++slot) {
    if (clazz->methods[slot] == method) {
	  break;
	}
  }

  /*
   * We use the NULL-ness of name_field to determine if we are
   * working with a Constructor or a Method
   */
  if (name_field != NULL) {
    jstring mname = (*env)->NewStringUTF(env, method->name);
    jclass returntype = sig_to_jclass(env,
        method->sig->method.return_type);
    (*env)->SetObjectField(env, rmethod, returnType_field,
			   returntype);
    mname = (*env)->NewGlobalRef(env, mname);
    (*env)->SetObjectField(env, rmethod, name_field, mname);
  }

  /* Create the declared exceptions array */
  exception_array = (*env)->NewObjectArray(env, num_exceptions,
      class_class, NULL);
  for (p = 0; p < num_exceptions; p ++)
  {
    jobject exc;
    ClazzFile *exc_type;
  
    exc_type = getThrowableException(env, method, p);
  
    exc = clazzfile_to_jclass(env, exc_type);
  
    (*env)->SetObjectArrayElement(env, exception_array, p, exc);
  }
  exception_array = (*env)->NewGlobalRef(env, exception_array);
  
  /* Create the parameter types array */
  param_array = (*env)->NewObjectArray(env, num_params,
      class_class, NULL);
  for (p = 0; p < num_params; p ++)
  {
    jclass param;
  
    param = sig_to_jclass(env, method->sig->method.params[p]);

    (*env)->SetObjectArrayElement(env, param_array, p, param);
  }
  param_array = (*env)->NewGlobalRef(env, param_array);
  
  (*env)->SetIntField(env, rmethod, slot_field, slot);
  (*env)->SetObjectField(env, rmethod, clazz_field,
			 clazzfile_to_jclass(env, clazz));
  (*env)->SetObjectField(env, rmethod, exceptionTypes_field,
			 exception_array);
  (*env)->SetObjectField(env, rmethod, parameterTypes_field,
			 param_array);
  return (*env)->NewGlobalRef(env, rmethod);
}

/**
 * Convert from a fieldID into a Java object of class
 * java.lang.reflect.Field.  The only difference with the JDK 1.2 JNI
 * ToReflectedField function is that by using this method,
 * repetitive string lookups can be avoided when inside loops.
 */
jobject
to_reflected_field(JNIEnv *env, jfieldID fieldID,
    jclass field_class, jmethodID field_ctor,
    jfieldID slot_field, jfieldID type_field, jfieldID name_field,
    jfieldID clazz_field)
{
  jobject rfield;
  jstring name;
  jclass type;
  FieldStruct *field = (FieldStruct *) fieldID;
  ClazzFile *clazz = field->clazz;
  int slot;

  /* Determine field's slot inside class */
  for (slot = 0; slot < clazz->num_fields; ++slot) {
    if (clazz->fields[slot] == field) {
	  break;
	}
  }
  rfield = (*env)->NewObject(env, field_class, field_ctor);
  name = (*env)->NewStringUTF(env, field->name);
  type = sig_to_jclass(env, field->sig);

  name = (*env)->NewGlobalRef(env, name);
  (*env)->SetObjectField(env, rfield, name_field, name);
  (*env)->SetIntField(env, rfield, slot_field, slot);
  (*env)->SetObjectField(env, rfield, clazz_field,
			 clazzfile_to_jclass(env, clazz));
  (*env)->SetObjectField(env, rfield, type_field, type);
  return (*env)->NewGlobalRef(env, rfield);
}
